;FAT16 driver layer
;FAT_Initialize
;FAT_getVolumeName
;FAT_getSectorOfCluster
;FAT_findEntry
;PartitionAdjust
;FAT_findFileInCluster
;FAT_getNextCluster
;FAT_nameConvertTo11
;FAT_nameConvertFrom11
;FAT_lookupPath
;DOS_openFile
;DOS_fileRead
;DOS_fileSeek
;DOS_countFilesInDir
;FAT_getEntryInfo
;DOS_getNextFile
;DOS_getFileSize_DirEntry
;DOS_getDirEntry
;File handle format:
fHandleLen EQU 22
;Flags:
flagSequential EQU 01h
flagEOF EQU 02h
;Offset:
;	0-3	32-bit size of file
;	4-5	16-bit starting cluster of file
;	6-7	16-bit current cluster offset for seek pointer
;	8		8-bit current sector offset for seek pointer (first sector of cluster in 7-8 would be 0)
;	9-12	32-bit seek offset (bytes from start of file)
;	13-14	16-bit current byte offset from start of seek sector
;	15-18	32-bit seek sector (adjusted for everything, including partition offset
;	19-20	16-bit current cluster of the seek pointer
;	21		8-bit flag byte
DOS_getFileSize_DirEntry:
	;Input - HL points to the directory entry for a file.
	;Output - DEHL is size (32-bit).
	ld de,28
	add hl,de
	ld e,(hl)
	inc hl
	ld d,(hl)
	inc hl
	ld a,(hl)
	inc hl
	ld h,(hl)
	ld l,a
	ex de,hl
	ret
DOS_getDirEntry:
	;Input - BC points to the current entry number of a file.
	;Input - HL points to the starting cluster of the directory.
	;Output - HL points to the directory entry of the file.
	push hl
	push bc
	call FAT_getEntryInfo
	pop bc
	ld (OP6+8),bc
	pop hl
	ld bc,(OP6+4)
	ld a,b
	or c
	jr z,$F
DOS_GDR_C_1:
	push bc
	call FAT_getNextCluster
	pop bc
	dec bc
	ld a,b
	or c
	jr nz,DOS_GDR_C_1
$$:
	ld (OP6+10),hl
	call FAT_getSectorOfCluster
	ld a,(OP6+3)
	ld c,a
	ld b,0
	add hl,bc
	jr nc,$F
	inc de
$$:
	;DEHL is sector...get byte offset
	call PartitionAdjust
	ld de,(SectorBuffer)
	push de
	call UFI_Read
	pop de
	ld hl,(OP6+1)
	add hl,de
	ret
DOS_getNextFile:
	;Returns the next file in a directory
	;Input - HL is the starting cluster of the directory.
	;Input - BC is the current entry number of a file.
	;Output - BC is the entry of the next file, or C if there are no more.
	;Output - HL points to the directory entry of the next file.
	;Destroys OP6.
	inc bc
	push hl
	push bc
	call FAT_getEntryInfo
	pop bc
	ld (OP6+8),bc
	pop hl
	ld bc,(OP6+4)
	ld a,b
	or c
	jr z,$F
DOS_GNF_C_1:
	push bc
	call FAT_getNextCluster
	pop bc
	dec bc
	ld a,b
	or c
	jr nz,DOS_GNF_C_1
$$:
	ld (OP6+10),hl
	call FAT_getSectorOfCluster
	ld a,(OP6+3)
	ld c,a
	ld b,0
	add hl,bc
	jr nc,$F
	inc de
$$:
	;DEHL is sector...get byte offset
	call PartitionAdjust
DOS_GNF_SectorLoop:
	ld de,(SectorBuffer)
	push de
	call UFI_Read
	ld de,(OP6+1)
	ld hl,(BPS)
	or a
	sbc hl,de
	;The remainder
	ld b,h
	ld c,l
	pop hl
	add hl,de
	srl b
	rr c
	srl b
	rr c
	srl b
	rr c
	srl b
	rr c
	srl b
	rr c
	ld b,c
DOS_GNF_Loop:
	ld a,(hl)
	or a
	jr z,DOS_GNF_NF
	cp 5
	jr z,DOS_GNF_SE
	cp 0E5h
	jr z,DOS_GNF_SE
	;We've found an entry!?
	push hl
	ld de,11
	add hl,de
	ld a,(hl)
	and 8			;It better not be a volume label or I will be very angry.
	pop hl
	jr nz,DOS_GNF_SE
	;Okay we have an entry
	or a
	ld bc,(OP6+8)
	ret		;Return with it!
DOS_GNF_SE:
	ld de,(OP6+8)
	inc de
	ld (OP6+8),de
	;Loop back for some more fun!
	ld de,32
	add hl,de
	djnz DOS_GNF_Loop
	;New sector now?
	ld a,(SPC)
	ld b,a
	ld a,(OP6+3)
	inc a
	cp b
	jr z,DOS_GNF_NC
	ld (OP6+3),a
	ld hl,LBA
	inc (hl)
	jr nz,$F
	inc hl
	inc (hl)
	jr nz,$F
	inc hl
	inc (hl)
	jr nz,$F
	inc hl
	inc (hl)
$$:
	ld hl,0
	ld (OP6+1),hl
	jr DOS_GNF_SectorLoop
DOS_GNF_NC:
	xor a
	ld (OP6+3),a
	;New cluster
	ld hl,(OP6+10)
	call FAT_getNextCluster
	ld (OP6+10),hl
	push hl
	dec hl
	dec hl
	ld bc,12h
	add hl,bc
	pop hl
	jr c,DOS_GNF_NF
	call FAT_getSectorOfCluster
	call PartitionAdjust
	jr DOS_GNF_SectorLoop
DOS_GNF_NF:
	scf
	ld bc,(OP6+8)
	ret
FAT_getEntryInfo:
	;Input - BC is the entry number. HL is the starting cluster of the directory.
	;Output - IX points to OP6
	;Destroys OP6, IX
	;Outputs to OP6.
	;Returns the offsets for bytes, sectors, clusters for the entry.
	ld ix,OP6
	ld (ix+6),l
	ld (ix+7),h
	;OP6+1 - 2 bytes, byte offset
	;OP6+3 - 1 byte, sector offset
	;OP6+4 - 2 bytes, cluster offset
	;OP6+6 - 2 bytes, starting cluster
	ld h,b
	ld l,c
	ld de,0
	ld bc,0
	add hl,hl
	rl e
	rl d
	add hl,hl
	rl e
	rl d
	add hl,hl
	rl e
	rl d
	add hl,hl
	rl e
	rl d
	add hl,hl
	rl e
	rl d
	;DEHL is the 32-bit offset of the file
	;Now find the sector offset
	ld bc,(BPS)
	dec bc
	ld a,b
	and h
	ld (ix+2),a
	ld a,c
	and l
	ld (ix+1),a
	;Divide by BPS
$$:
	srl d
	rr e
	rr h
	rr l
	srl b
	rr c
	jr nz,$B
	;DEHL is now a sector offset
	ld a,(SPC)
	dec a
	and l
	ld (ix+3),a
	;Divide by SPC
$$:
	srl d
	rr e
	rr h
	rr l
	rra
	jr nz,$B
	;
	ld (ix+4),l
	ld (ix+5),h		;HL is the cluster of the offset
	ret
DOS_countFilesInDir:
	;Counts the directory entries in a directory
	;Input: HL - path of directory, ASCIIZ string
	;Output - BC is the number of entries
	;Destroys OP6.
	inc hl
	ld a,(hl)
	dec hl
	or a
	jr nz,$F
	ld a,(hl)
	cp '/'
	jr nz,$F
	ld hl,0
	jr DOS_CFID_IP1
$$:
	call FAT_lookupPath
	ld bc,26
	add hl,bc
	ld a,(hl)
	inc hl
	ld h,(hl)
	ld l,a
DOS_CFID_IP1:
	;HL is cluster
	xor a
	ld (OP6),a
	;(OP6) is sector offset
	ld bc,0
	ld (OP6+1),bc		;OP6+1 is the count of files found
DOS_CFID_Loop:
	push hl
	call FAT_getSectorOfCluster
	call PartitionAdjust
DOS_CFID_SectorLoop:
	ld de,(SectorBuffer)
	call UFI_Read
	ld hl,(SectorBuffer)
	ld bc,(BPS)
	srl b
	rr c
	srl b
	rr c
	srl b
	rr c
	srl b
	rr c
	srl b
	rr c
	ld b,c
DOS_CFID_S_Loop:
	ld a,(hl)
	or a
	jr z,DOS_CFID_Done
	cp 5		;E5h in the Kanji character set
	jr z,DOS_CFID_SE
	cp 0E5h
	jr z,DOS_CFID_SE
	push hl
	ld de,11
	add hl,de
	ld a,(hl)
	and 8
	pop hl
	jr nz,DOS_CFID_SE
	;We have a valid directory entry
	ld de,(OP6+1)
	inc de
	ld (OP6+1),de
DOS_CFID_SE:
	ld de,32
	add hl,de
	djnz DOS_CFID_S_Loop
	;Done with a sector. Next one?
	ld a,(SPC)
	ld b,a
	ld a,(OP6)
	inc a
	cp b
	jr z,DOS_CFID_NC
	ld (OP6),a
	ld hl,LBA
	inc (hl)
	jr nz,$F
	inc hl
	inc (hl)
	jr nz,$F
	inc hl
	inc (hl)
	jr nz,$F
	inc hl
	inc (hl)
$$:
	jr DOS_CFID_SectorLoop
DOS_CFID_NC:
	;go to next cluster if available
	xor a
	ld (OP6),a
	pop hl
	call FAT_getNextCluster
	push hl
	dec hl
	dec hl
	ld bc,12h
	add hl,bc		;C if out of range
	jr c,DOS_CFID_Done
	pop hl
	jr DOS_CFID_Loop
DOS_CFID_Done:
	pop hl
	ld bc,(OP6+1)
	or a
	ret
DOS_fileSeek:
	;Positions the seek pointer in a file.
	;Input - IX is a pointer to the file handle of an opened file.
	;Input - DEHL is the offset to move the seek pointer too
	ld (ix+9),l
	ld (ix+10),h
	ld (ix+11),e
	ld (ix+12),d
	call DOS_OptimizeFH
	or a
	ret
DOS_fileRead:
	;Reads bytes from a file.
	;Input - HL pointer to file handle of opened file. DE is where to store bytes read.
	;Input - BC is the number of bytes to read.
	;Output - Reads bytes. Returns C if it failed or encountered EOF first. BC is the number
	;of bytes returned.
	;Destroys OP6
	;OP6+0 - Location to store next bytes read
	;OP6+2 - Number of bytes read so far
	;OP6+4 - Number of bytes remaining to read
	push hl
	pop ix
	ld (ix+21),0		;clear flags
	ld hl,0
	ld (OP6+2),hl
	ld (OP6),de
	;IX is the pointer to the file handle, for DOS_OptimizeFH later
	;Check how many bytes we are reading...is it too much?
	ld l,(ix+0)
	ld h,(ix+1)
	ld e,(ix+2)
	ld d,(ix+3)
	push bc
	or a
	sbc hl,bc
	ex de,hl
	ld bc,0
	sbc hl,bc
	ex de,hl
	pop bc
	;DEHL is the size of the file minus what we are reading
	;C? C is for cookie and that's good enough for me...
	jr nc,$F
	;Cookie cookie cookie starts with C!
	ld c,(ix+0)
	ld b,(ix+1)
	ld a,flagEOF
	or (ix+21)
	ld (ix+21),a
$$:
	ld (OP6+4),bc
DOS_FR_Loop:
	;Shall we read some bytes? Sure!
	ld bc,(OP6+4)
	ld hl,(BPS)
	ld e,(ix+13)
	ld d,(ix+14)
	or a
	sbc hl,de
	;HL is the most bytes we can read at once
	sbc hl,bc
	;If BC > HL, we have a problem
	jr nc,$F
	add hl,bc
	ld b,h
	ld c,l		;We will only read BC
$$:
	;BC is the number of bytes to read
	push bc
	push de
	ld l,(ix+15)
	ld h,(ix+16)
	ld (LBA),hl
	ld l,(ix+17)
	ld h,(ix+18)
	ld (LBA+2),hl
	ld de,(SectorBuffer)
	push ix
	call UFI_Read
	pop ix
	pop de
	ld hl,(SectorBuffer)
	add hl,de
	;HL is the offset...now copy BC (on stack) bytes
	pop bc
	push bc
	ld de,(OP6+0)
	ldir
	ld (OP6+0),de		;new location to store the next few bytes
	pop bc
	ld hl,(OP6+2)
	add hl,bc
	ld (OP6+2),hl
	ld hl,(OP6+4)
	or a
	sbc hl,bc
	ld (OP6+4),hl
	;Now we update pointers and things
	ld l,(ix+9)
	ld h,(ix+10)
	ld e,(ix+11)
	ld d,(ix+12)
	add hl,bc
	jr nc,$F
	inc de
$$:
	ld (ix+9),l
	ld (ix+10),h
	ld (ix+11),e
	ld (ix+12),d
	ld l,(ix+13)
	ld h,(ix+14)
	add hl,bc
	ld (ix+13),l
	ld (ix+14),h
	;Are we at a sector boundary?
	ld bc,(BPS)
	dec bc
	ld a,h
	and b
	ld b,a
	ld a,l
	and c
	or b		;Z if we are
	jr nz,DOS_FR_PreLoop
	ld (ix+13),0
	ld (ix+14),0
	;Increment sector counts
	inc (ix+15)
	jr nz,DOS_FR_2
	inc (ix+16)
	jr nz,DOS_FR_2
	inc (ix+17)
	jr nz,DOS_FR_2
	inc (ix+18)
DOS_FR_2:
	ld a,(SPC)
	ld b,a
	ld a,(ix+8)
	inc a
	cp b
	jr z,DOS_FR_NextCluster
	ld (ix+8),a
DOS_FR_PreLoop:
	ld bc,(OP6+4)
	ld a,b
	or c
	jr nz,DOS_FR_Loop
	ld bc,(OP6+2)
	ld a,(ix+21)
	and flagEOF
	ret z
	scf
	ret
DOS_FR_NextCluster:
	xor a
	ld (ix+8),a
	inc (ix+6)
	jr nz,$F
	inc (ix+7)
$$:
	ld l,(ix+19)
	ld h,(ix+20)
	push ix
	call FAT_getNextCluster
	pop ix
	push ix
	ld (ix+19),l
	ld (ix+20),h
	call FAT_getSectorOfCluster
	pop ix
	ld bc,(PartStart)
	add hl,bc
	ex de,hl
	ld bc,(PartStart+2)
	adc hl,bc
	ex de,hl
	ld (ix+15),l
	ld (ix+16),h
	ld (ix+17),e
	ld (ix+18),d
	jr DOS_FR_PreLoop
DOS_openFile:
	;Opens a file for reading or writing. Populates a file handle pointing to important
	;information about the open file.
	;Input - HL points to an ASCIIZ string containing a filename
	;Input - DE points to a location to store the file handle.
	;Output - HL points to file handle. Returns C if failure.
	push de
	call FAT_lookupPath
	pop de
	ret c
	;Check if directory instead of file
	ld bc,11
	add hl,bc
	ld a,(hl)
	and 18h
	scf
	ret nz		;if NZ then it was either a volume label or a directory, NOT A FILE! EVIL USER!
	push de
	ld bc,28-11
	add hl,bc
	ld bc,4
	ldir
	ld bc,-6
	add hl,bc
	ldi
	ldi
	;Initialize seek to 0
	ld h,d
	ld l,e
	ld (hl),0
	inc de
	ld bc,6
	ldir
	pop hl
	push hl
	push hl
	pop ix
	ld (ix+21),0
	call DOS_OptimizeFH
	pop hl
	or a
	ret
DOS_OptimizeFH:
	;Routine to adjust seek pointers in a file handle whenever the seek position is changed.
	;Input - IX points to start of file handle
	ld l,(ix+9)
	ld h,(ix+10)
	ld e,(ix+11)
	ld d,(ix+12)
	;DEHL is the 32-bit offset of the file
	;Now find the sector offset
	ld bc,(BPS)
	dec bc
	ld a,b
	and h
	ld (ix+14),a
	ld a,c
	and l
	ld (ix+13),a
	;Divide by BPS
DOS_OFH1:
	srl d
	rr e
	rr h
	rr l
	srl b
	rr c
	jr nz,DOS_OFH1
	;DEHL is now a sector offset
	ld a,(SPC)
	dec a
	and l
	ld (ix+8),a
	;Divide by SPC (Sectors per cluster, or seconds per character if you're a really slow typist)
DOS_OFH2:
	srl d
	rr e
	rr h
	rr l
	rra
	jr nz,DOS_OFH2
	;
	ld (ix+6),l
	ld (ix+7),h		;HL is the cluster of the offset
	ld b,h
	ld c,l
	ld a,(ix+21)
	and flagSequential
	jr z,$F
	ld l,(ix+19)
	ld h,(ix+20)
	call FAT_getNextCluster
	jr DOS_OFH4
$$:
	ld l,(ix+4)
	ld h,(ix+5)		;starting cluster
	ld a,b
	or c
	jr z,DOS_OFH4
DOS_OFH3:
	;Loop through!
	push bc
	call FAT_getNextCluster
	pop bc
	dec bc
	ld a,b
	or c
	jr nz,DOS_OFH3
DOS_OFH4:
	ld (ix+19),l
	ld (ix+20),h
	;Now lookup the sector for it
	call FAT_getSectorOfCluster
	;DEHL is the sector
	;Add our sector offset
	ld c,(ix+8)
	ld b,0
	add hl,bc
	ld bc,0
	ex de,hl
	adc hl,bc
	ex de,hl
	;We have a sector! Adjust for partition offset
	ld bc,(PartStart)
	add hl,bc
	ex de,hl
	ld bc,(PartStart+2)
	adc hl,bc
	ex de,hl
	ld (ix+15),l
	ld (ix+16),h
	ld (ix+17),e
	ld (ix+18),d
	ret
FAT_lookupPath:
	;Input - HL points to an ASCIIZ string containing a path to a folder/file.
	;Output - HL points to the directory entry for that. C if can't find.
	;Destroys OP6
	ld a,(hl)
	cp '/'
	scf
	ret nz
	inc hl
	;Set root cluster
	ld de,0
FAT_lookupPath_Loop:
	ld a,(hl)
	or a
	scf
	ret z
	push hl
	push de
	ld de,OP6
	call FAT_nameConvertTo11		;Take one part of the name...
	pop hl
	ld de,OP6
	call FAT_findFileInCluster
	;C set if we couldn't find it. Otherwise HL points to the FAT directory entry.
	push hl
	pop ix			;save it just in case
	jr c,FAT_lookupPath_C		;couldn't find it!
	ld de,26
	add hl,de
	ld e,(hl)
	inc hl
	ld d,(hl)
	;DE is the first cluster of this file/directory
	;Now is there another level to search or are we at the end?
	pop hl
	;If we just did "FOLDER1" out of "/FOLDER2/FOLDER1/FOLDER0", then HL points to
	;the "F" in "FOLDER1".
FAT_lookupPath_1:
	ld a,(hl)
	inc hl
	cp '/'
	jr z,FAT_lookupPath_Loop
	or a
	jr nz,FAT_lookupPath_1
	;We couldn't find anything else. We must be done. HURRAY!
	push ix
	pop hl
	or a
	ret
FAT_lookupPath_C:
	pop hl
	ret
FAT_nameConvertFrom11:
	;Converts a filename like "BLAH    TXT" into a regular formatted name.
	;Input - HL points to filename, DE points to 13 byte buffer (at most).
	;Output - HL points to formatted name
	push de
	ld b,8
	call FAT_NCFES
	ld a,'.'
	ld (de),a
	inc de
	ld b,3
	call FAT_NCFES
	;Check if there was no extension...then remove period
	dec de
	ld a,(de)
	cp '.'
	jr z,FAT_NCFES_1
	inc de
FAT_NCFES_1:
	xor a
	ld (de),a
	pop hl
	ret
FAT_NCFES:
	ld a,(hl)
	cp ' '
	jr z,$F
	ld (de),a
	inc de
$$:
	inc hl
	djnz FAT_NCFES
	ret
FAT_nameConvertTo11:
	;Converts a filename like "dummy.txt" into "DUMMY   TXT" (11 character format)
	;Input - HL points to filename, DE points to 11 byte buffer for formatted name (OP
	;			registers work particularly well for this). HL can be ASCIIZ string or
	;terminated with a forward slash.
	;Output - HL points to formatted name
	push de
	ld b,8
	call FAT_nameConvert_Subroutine
	ld b,3
	call FAT_nameConvert_Subroutine
	pop hl
	ret
FAT_nameConvert_Subroutine:
	;Input - HL is source. DE is destination. B is the number of bytes to fill.
	ld a,(hl)
	cp '.'
	jr z,FAT_NC_1
	cp '/'
	jr z,FAT_NC_1
	or a
	jr z,FAT_NC_1
	cp 'a'
	jr c,$F
	cp 'z'+1
	jr nc,$F
	add a,'A'-'a'
$$:
	ld (de),a
	inc hl
	inc de
	djnz FAT_nameConvert_Subroutine
	ret
FAT_NC_1:
	ld a,' '
	ld (de),a
	inc de
	djnz FAT_NC_1
	ld a,(hl)
	cp '.'
	ret nz
	inc hl
	ret
FAT_getNextCluster:
	;Input - HL is a cluster.
	;Output - HL is the cluster following this one (literally the WORD in the FAT for the
	;			 input cluster).
	;First we must calculate the sector/offset.
	ld de,(BPS)
	xor a
	add hl,hl		;AHL = 2HL
	adc a,a
	dec de
	push af
	ld a,h
	and d
	ld b,a
	ld a,l
	and e
	ld c,a		;BC = (2AHL) MOD DE
	pop af
FAT_getNextCluster_1:
	sra a
	rr h
	rr l
	srl d
	rr e
	jr nz,FAT_getNextCluster_1
	;HL is the sector...BC is the offset
	ld de,(FATStart)
	add hl,de
	ld de,0
	call PartitionAdjust
	push bc
	ld de,(SectorBuffer)
	push de
	call UFI_Read
	pop de
	pop hl
	add hl,de
	ld a,(hl)
	inc hl
	ld h,(hl)
	ld l,a
	ret
FAT_findFileInCluster:
	;Input - HL is a cluster.
	;Input - DE points to the name of a file (directory). 11 characters.
	;Output - C set if it couldn't find it. Otherwise HL points to the directory entry.
FAT_FFIC_Loop:
	push de
	push hl
	call FAT_findEntry		;Look in cluster HL
	jr nc,FAT_FFIC_Found
	pop hl
	pop de
	;Move to next cluster?
	push de
	call FAT_getNextCluster
	pop de
	;HL is the next cluster
	push hl
	dec hl
	dec hl
	ld bc,12h
	add hl,bc		;C if out of range
	pop hl
	ret c
	jr FAT_FFIC_Loop
FAT_FFIC_Found:
	pop de
	pop de
	ret
PartitionAdjust:
	;Input - DEHL is a sector offset from the volume start
	;Output - loads (LBA) 4 bytes to compensate for the partition location
	push bc
	ld bc,(PartStart)
	add hl,bc
	ld (LBA),hl
	ex de,hl
	ld bc,(PartStart+2)
	adc hl,bc
	ld (LBA+2),hl
	ex de,hl
	pop bc
	ret
FAT_findEntry:
	;Finds an entry in a cluster. This is useful if you looked up the cluster
	;for a directory and want to find a file in that directory.
	;Inputs - HL is a cluster number.
	;DE points to the name of a file (directory), 11 characters
	;Output - C if could not find it. Otherwise HL points to the start of the
	;directory entry.
	push de
	call FAT_getSectorOfCluster
	;DEHL is a sector
	call PartitionAdjust
	ld a,(SPC)
	ld b,a
	pop de
FAT_findEntry_SectorLoop:
	push bc
	push de
	ld de,(SectorBuffer)
	call UFI_Read
	pop de
	ld hl,(LBA)
	ld bc,1
	add hl,bc
	ld (LBA),hl
	ld hl,(LBA+2)
	ld bc,0
	adc hl,bc
	ld (LBA+2),hl
	ld hl,(SectorBuffer)
	ld bc,(BPS)
	srl b
	rr c
	srl b
	rr c
	srl b
	rr c
	srl b
	rr c
	srl b
	rr c			;Divide by 32, the number of entries
FAT_findEntry_EntryLoop:
	push bc
	;HL is the entry
	;DE is our string
	;Compare
	push de
	push hl
	ld bc,11
	ld a,(hl)
	or a
	jr z,FAT_findEntry_LastOne
FAT_findEntry_NameCheck:
	ld a,(de)
	inc de
	cpi
	jr nz,FAT_findEntry_SkipAhead
	jp pe,FAT_findEntry_NameCheck
	;Match!
	pop hl
	pop de
	pop bc
	pop bc
	or a
	ret
FAT_findEntry_SkipAhead:
	pop hl
	pop de
	ld bc,32
	add hl,bc
	pop bc
	dec bc
	ld a,b
	or c
	jr nz,FAT_findEntry_EntryLoop
	pop bc
	djnz FAT_findEntry_SectorLoop
	scf
	ret
FAT_findEntry_LastOne:
	pop hl
	pop de
	pop bc
	pop bc
	scf
	ret
FAT_getSectorOfCluster:
	;Inputs - HL is cluster N
	;Outputs - DEHL is the first sector of that cluster
	ld a,h
	or l
	jr z,FAT_getSectorOfCluster_Root
	dec hl
	dec hl
	;DEHL = DEHL - 2
	ld de,0
	;Multiply by SPC
	ld a,(SPC)
	ld b,-1
FAT_getSectorOfCluster_1:
	inc b
	rra
	jr nc,FAT_getSectorOfCluster_1
	ld a,b
	or a
	jr z,$F
	;Multiply!
FAT_getSectorOfCluster_2:
	add hl,hl
	rl e
	rl d
	djnz FAT_getSectorOfCluster_2
$$:
	;Add FirstDataSector
	ld bc,(FirstDataSector)
	add hl,bc
	ex de,hl
	ld bc,(FirstDataSector+2)
	adc hl,bc
	ex de,hl
	ret
FAT_getSectorOfCluster_Root:
	ld de,0
	ld hl,(FirstRootDirSecNum)
	ret
FAT_getVolumeName:
	;Inputs: None
	;Output - HL points to an 11-byte string containing the volume name
	ld hl,(PartStart)
	ld (LBA),hl
	ld hl,(PartStart+2)
	ld (LBA+2),hl
	ld de,(SectorBuffer)
	push de
	call UFI_Read
	pop hl
	ld de,43
	add hl,de
	ret
FAT_Initialize:
	ld hl,(PartStart)
	ld (LBA),hl
	ld hl,(PartStart+2)
	ld (LBA+2),hl
	ld de,(SectorBuffer)
	push de
	call UFI_Read
	pop hl
	ld de,13
	add hl,de
	ld a,(hl)
	ld (SPC),a		;Get the number of sectors per cluster
	ld ix,(SectorBuffer)
	ld e,(ix+19)
	ld d,(ix+20)
	;16-bit count
	ld (FATSectors),de
	;Now we have gotten the total sector count
	ld hl,(SectorBuffer)
	ld de,510
	add hl,de
	ld e,(hl)
	inc hl
	ld d,(hl)
	ld hl,0AA55h
	or a
	sbc hl,de
	ld a,errIBS
	jp nz,displayErr
	;Compute RootDirSectors
	;get BPB_RootEntCnt
	ld ix,(SectorBuffer)
	ld l,(ix+17)
	ld h,(ix+18)
	;HL is BPB_RootEntCnt
	;RootDirSectors = ((BPB_RootEntCnt * 32) + (BPB_BytesPerSec - 1)) / BPB_BytesPerSec;
	add hl,hl	;x2
	add hl,hl	;x4
	add hl,hl	;x8
	add hl,hl	;x16
	add hl,hl	;x32
	ld de,(BPS)
	dec de
	add hl,de
	B_CALL SetXXXXOP2
	B_CALL OP2ToOP1
	ld hl,(BPS)
	B_CALL SetXXXXOP2
	B_CALL FPDiv
	B_CALL CkPosInt
	jr z,$F
	B_CALL Plus1
$$:
	B_CALL Trunc
	call ConvOP1
	ld (RootDirSectors),de
	push de
	;FirstDataSector = BPB_ResvdSecCnt + (BPB_NumFATs * BPB_FATSz16) + RootDirSectors
	ld ix,(SectorBuffer)
	ld a,(ix+16)
	push ix
	B_CALL SetXXOP1
	pop ix
	ld l,(ix+22)
	ld h,(ix+23)
	push ix
	B_CALL SetXXXXOP2
	B_CALL FPMult
	call ConvOP1
	pop ix
	pop hl
	add hl,de
	ld e,(ix+14)
	ld d,(ix+15)
	ld (FATStart),de
	add hl,de
	dec hl			;I don't know why we have to do this. Random magic?
	ld (FirstDataSector),hl
	ld de,(RootDirSectors)
	or a
	sbc hl,de
	inc hl			;More random magic
	ld (FirstRootDirSecNum),hl
	ret
ConvOP1:
	;Gets a number out of OP1
	;Returns C if the number was too large
	ld b,5
	ld de,0
ConvOP1_Loop:
	push bc
	xor a
	ld hl,OP1+4
	rld
	dec hl
	rld
	dec hl
	rld
	ex de,hl
	add hl,hl
	push hl
	add hl,hl
	add hl,hl
	jr c,ConvOP1_Return2
	pop bc
	add hl,bc
	jr c,ConvOP1_Return
	;HL is times 10
	ld b,0
	ld c,a
	add hl,bc
	ex de,hl
	dec hl
	pop bc
	dec (hl)
	ret p
	djnz ConvOP1_Loop
	scf
	ret
ConvOP1_Return2:
	pop bc
ConvOP1_Return:
	pop bc
	ret
